import os
import rasterio
from .util import run_command

from osgeo import gdal
# Fix a bug in conda gdal: https://github.com/OSGeo/gdal/issues/1231
if ';' in os.environ["PATH"]:
    os.environ["PATH"] = os.environ["PATH"].split(';')[1]


def stack_bands(files, outfile, buildvrt=False):
    if not buildvrt:
        nodata = rasterio.open(files[0]).nodata
        if nodata is not None:
            args = ['gdal_merge.py', '-separate', '-a_nodata', nodata, '-o', outfile, *files]
            run_command(args)
        else:
            args = ['gdal_merge.py', '-separate', '-o', outfile, *files]
            run_command(args)
    else:
        args = ['gdalbuildvrt', '-separate', '-overwrite', outfile, *files]
        run_command(args)

    # # Copy existing band descriptions to the merged image
    # descriptions = ()
    # for file in files:
    #     descriptions += rasterio.open(file).descriptions
    # ds = gdal.Open(outfile, gdal.GA_Update)
    # for band, desc in enumerate(descriptions, start=1):
    #     if desc is not None:
    #         rb = ds.GetRasterBand(band)
    #         rb.SetDescription(desc)
    # del ds

    return outfile


def extract_bands(file, bands=None):
    dataset = rasterio.open(file)
    if bands is None:
        bands = range(1, dataset.count + 1)
    elif isinstance(bands, int):
        bands = [bands]
    meta = dataset.meta.copy()
    meta.update({
        'count': 1
    })
    outfiles = []
    descriptions = rasterio.open(file).descriptions
    for band in bands:
        outfile = os.path.splitext(file)[0] + "_band_{}".format(band) + os.path.splitext(file)[1]
        with rasterio.open(outfile, 'w', **meta) as dst:
            dst.write(dataset.read(band), 1)
            if descriptions[band - 1] is not None:
                dst.set_band_description(1, descriptions[band-1])
        outfiles.append(outfile)
    return outfiles



import rasterio
from rasterio.plot import show

# https://rasterio.readthedocs.io/en/latest/api/rasterio.plot.html?highlight=show#rasterio.plot.show
# https://matplotlib.org/api/_as_gen/matplotlib.axes.Axes.imshow.html#matplotlib-axes-axes-imshow


def plot(file, **kwargs):
    with rasterio.open(file) as dataset:
        show(dataset, **kwargs)



import os
import rasterio
from rasterio.windows import Window, transform
from rasterio.io import MemoryFile
from .util import run_command


def get_bounds(file):
    if isinstance(file, bytes):
        return MemoryFile(file).open().bounds
    return rasterio.open(file).bounds


def get_shape(file):
    if isinstance(file, bytes):
        return MemoryFile(file).open().shape
    return rasterio.open(file).shape


def rescale(infile, outfile, src_min, src_max, dst_min=0, dst_max=255, dtype=None):
    args = ['gdal_translate']
    if dtype is not None:
        args.extend(['-ot', dtype])
    args.extend(['-scale', src_min, src_max, dst_min, dst_max, infile, outfile])
    run_command(args)
    return outfile


def merge(files, outfile, bbox=None, init_val=None, ignore_val=None, nodata_val=None):
    try:
        os.remove(outfile)
    except OSError:
        pass
    args = ['gdal_merge.py', '-o', outfile]
    if init_val is not None:
        args.extend(['-init', init_val])
    if ignore_val is not None:
        args.extend(['-n', ignore_val])
    if nodata_val is not None:
        args.extend(['-a_nodata', nodata_val])
    else:
        nodata = rasterio.open(files[0]).nodata
        if nodata is not None:
            args.extend(['-a_nodata', nodata])
    if bbox is not None:
        left, bottom, right, top = bbox
        args.extend(['-ul_lr', left, top, right, bottom])
    args.extend(files)
    if not os.path.exists(os.path.dirname(os.path.abspath(outfile))):
        os.makedirs(os.path.dirname(os.path.abspath(outfile)))
    run_command(args)

    # # Copy existing band descriptions to the merged image
    # from osgeo import gdal
    # # Fix a bug in conda gdal: https://github.com/OSGeo/gdal/issues/1231
    # if ';' in os.environ["PATH"]:
    #     os.environ["PATH"] = os.environ["PATH"].split(';')[1]
    # descriptions = rasterio.open(files[0]).descriptions
    # ds = gdal.Open(outfile, gdal.GA_Update)
    # for band, desc in enumerate(descriptions, start=1):
    #     if desc is not None:
    #         rb = ds.GetRasterBand(band)
    #         rb.SetDescription(desc)
    # del ds

    return outfile


def split(infile, width, height=None):
    dataset = rasterio.open(infile)
    height = height or width
    raw_width = dataset.width
    raw_height = dataset.height
    raw_window = Window(0, 0, raw_width, raw_height)
    filename = dataset.name
    outfiles = []
    for col_off in range(0, raw_width, width):
        for row_off in range(0, raw_height, height):
            window = Window(col_off, row_off, width, height).intersection(raw_window)
            meta = dataset.meta.copy()
            meta.update({
                'width': window.width,
                'height': window.height,
                'transform': transform(window, dataset.transform)
            })
            outfile = os.path.splitext(filename)[0] + "_{}_{}_{}_{}".format(col_off, row_off, window.width, window.height) + os.path.splitext(filename)[1]
            with rasterio.open(outfile, 'w', **meta) as dst:
                dst.write(dataset.read(window=window))
                # Copy existing band descriptions to the splitted images
                descriptions = rasterio.open(infile).descriptions
                for band, desc in enumerate(descriptions, start=1):
                    if desc is not None:
                        dst.set_band_description(band, desc)
                outfiles.append(outfile)
    return outfiles


def query(infile, bbox, outfile=None):
    dataset = rasterio.open(infile)
    left, bottom, right, top = bbox
    start = dataset.index(left, top)
    stop = dataset.index(right, bottom)
    window = Window.from_slices((start[0], stop[0]), (start[1], stop[1]), boundless=True)
    meta = dataset.meta.copy()
    meta.update({
        'width': window.width,
        'height': window.height,
        'transform': transform(window, dataset.transform)
    })
    if outfile is None:
        filename = dataset.name
        outfile = os.path.splitext(filename)[0] + "_{}_{}_{}_{}".format(left, bottom, right, top) + os.path.splitext(filename)[1]
    with rasterio.open(outfile, 'w', **meta) as dst:
        dst.write(dataset.read(window=window, boundless=True))
        # Copy existing band descriptions to the splitted images
        descriptions = rasterio.open(infile).descriptions
        for band, desc in enumerate(descriptions, start=1):
            if desc is not None:
                dst.set_band_description(band, desc)
    return outfile



import os
import xml.etree.ElementTree as ET
import subprocess


def run_command(args):
    args = [str(arg) for arg in args]
    try:
        subprocess.run(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=True)
    except subprocess.CalledProcessError as e:
        raise ValueError("""
            Command Failed: {}
            
            STDOUT: {}
            
            STDERR: {}
            """.format(' '.join(e.cmd), e.stdout.decode(), e.stderr.decode()))


def create_xyz_vrt(infile, lat_col='lat', lon_col='lon'):
    layer_name = os.path.splitext(os.path.basename(infile))[0]

    data_source = ET.Element('OGRVRTDataSource')
    layer = ET.SubElement(data_source, 'OGRVRTLayer', name=layer_name)
    ET.SubElement(layer, 'SrcDataSource').text = infile
    ET.SubElement(layer, 'GeometryType').text = 'wkbPoint'
    ET.SubElement(layer, 'LayerSRS').text = 'EPSG:4326'
    ET.SubElement(layer, 'GeometryField', encoding='PointFromColumns', x=lon_col, y=lat_col)
    tree = ET.ElementTree(data_source)
    return tree


def add_latitude(latitude, delta):
    return max(-90, min(90, latitude + delta))


def add_longitude(longitude, delta):
    result = longitude + delta
    while result > 180:
        result -= 360
    while result < -180:
        result += 360
    return result


def diff_longitude(start_longitude, end_longitude):
    diff = end_longitude - start_longitude
    if diff < 0:
        diff += 360
    return diff


def expand_bbox(bbox, delta):
    if diff_longitude(bbox[0], bbox[2]) + 2 * delta > 360:
        return -180,\
               add_latitude(bbox[1], -delta),\
               180,\
               add_latitude(bbox[3], delta)
    else:
        return add_longitude(bbox[0], -delta),\
               add_latitude(bbox[1], -delta),\
               add_longitude(bbox[2], delta),\
               add_latitude(bbox[3], delta)



import string
from .util import run_command


def calculate(infiles, outfile, expr, nodata_val=None, dtype=None, all_bands=None):
    args = ['gdal_calc.py']
    asciis = string.ascii_uppercase
    for idx, infile in enumerate(infiles):
        args.extend(['-{}'.format(asciis[idx]), infile])
    args.append('--outfile={}'.format(outfile))
    args.append('--calc={}'.format(expr))
    if nodata_val is not None:
        args.append('--NoDataValue={}'.format(nodata_val))
    if dtype is not None:
        args.append('--type={}'.format(dtype))
    if all_bands is not None:
        args.append('--allBands={}'.format(all_bands))
    args.append('--overwrite')
    run_command(args)
    return outfile



import os
from .util import run_command


def fill_nodata(infile, outfile=None, inplace=True, max_distance=None):
    if outfile is not None and inplace:
        raise ValueError('outfile cannot be set with inplace=True. When inplace=True, interpolation will be performed '
                         'in place on the source file and no new file will be generated. Set inplace=False and specify '
                         'outfile if you need a new file for the interpolated result')
    args = ['gdal_fillnodata.py']
    if max_distance is not None:
        args.extend(['-md', max_distance])
    if inplace:
        args.append(infile)
    else:
        outfile = outfile or os.path.splitext(infile)[0] + '_filled' + os.path.splitext(infile)[1]
        args.extend([infile, outfile])
    run_command(args)
    return infile if inplace else outfile



import os
import tempfile
from pyrip.util import expand_bbox


def download(cos_client, bucket, key, download_dir):
    if not os.path.exists(download_dir):
        os.makedirs(download_dir)
    cos_client.download_file(bucket, key, os.path.join(download_dir, os.path.basename(key)))
    return os.path.join(download_dir, os.path.basename(key))


def query_tiles_with_spark(spark, cos_api_key, cos_instance_crn, cos_endpoint_url, bucket, image_meta_prefix, layers,
                           start_date, end_date, bbox, download_dir, search_buffer_in_degrees=2, selected_dates=None,
                           merge_tiles=True, merge_by_date=True, force_bbox=True):

    if isinstance(layers, str):
        layers = [layers]

    if isinstance(selected_dates, str) or isinstance(selected_dates, int):
        selected_dates = [selected_dates]

    # Get COS client for downloading tiles
    import ibm_boto3
    from ibm_botocore.client import Config
    cos_client = ibm_boto3.client("s3",
        ibm_api_key_id=cos_api_key,
        ibm_service_instance_id=cos_instance_crn,
        endpoint_url=cos_endpoint_url,
        config=Config(signature_version="oauth")
    )

    # Config Spark for reading meta data
    sc = spark.sparkContext
    sc._jsc.hadoopConfiguration().set('fs.cos.myCos.iam.api.key', cos_api_key)
    sc._jsc.hadoopConfiguration().set('fs.cos.myCos.iam.service.id', cos_instance_crn)
    sc._jsc.hadoopConfiguration().set('fs.cos.myCos.endpoint', cos_endpoint_url)

    # Get meta data
    spark.read.parquet('cos://{}.myCos/{}'.format(bucket, image_meta_prefix)).createOrReplaceTempView('temp_view')

    expanded_bbox = expand_bbox(bbox, search_buffer_in_degrees)

    if selected_dates is None:
        date_query_part = 'date between {} AND {}'.format(start_date, end_date)
    else:
        date_query_part = 'date in {}'.format(tuple(selected_dates))

    if expanded_bbox[2] >= expanded_bbox[0]:
        lon_query_part = 'lon between {} AND {}'.format(expanded_bbox[0], expanded_bbox[2])
    else:
        lon_query_part = '(lon between {} AND 180 OR lon between -180 AND {})'.format(expanded_bbox[0], expanded_bbox[2])

    df = spark.sql("""
    SELECT *
    FROM temp_view
    WHERE layer in {0}
    AND {1}
    AND lat between {2[1]} AND {2[3]}
    AND {3}
    """.format(tuple(layers), date_query_part, expanded_bbox, lon_query_part)).toPandas()

    if not os.path.exists(download_dir):
        os.makedirs(download_dir)

    if merge_tiles:
        from pyrip.image import merge
        merged_images = []
        for layer in layers:
            tmp_df = df[df['layer'] == layer].sort_values(['date', 'lat', 'lon'])
            if merge_by_date:
                dates = tmp_df['date'].unique()
                for date in dates:
                    tiles = []
                    with tempfile.TemporaryDirectory() as tmpdir:
                        for url in tmp_df[tmp_df['date'] == date]['url']:
                            bucket, key = url.replace('cos://', '').split('/', 1)
                            tile = download(cos_client, bucket, key, tmpdir)
                            tiles.append(tile)
                        if bbox[2] > bbox[0]:
                            merged_image = os.path.join(download_dir, layer, str(date), 'merged.tif')
                            if force_bbox:
                                merge(tiles, merged_image, bbox)
                            else:
                                merge(tiles, merged_image)
                            merged_images.append(merged_image)
                        # If the given bbox spans over anti-meridian, we have to merge into two parts. Note that this
                        # will also overwrite the value of force_bbox and set it to true.
                        else:
                            merged_image_1 = os.path.join(download_dir, layer, str(date), 'merged_1.tif')
                            merged_image_2 = os.path.join(download_dir, layer, str(date), 'merged_2.tif')
                            merge(tiles, merged_image_1, (bbox[0], bbox[1], 180, bbox[3]))
                            merge(tiles, merged_image_2, (-180, bbox[1], bbox[2], bbox[3]))
                            merged_images.extend([merged_image_1, merged_image_2])
            else:
                tiles = []
                with tempfile.TemporaryDirectory() as tmpdir:
                    for url in tmp_df['url']:
                        bucket, key = url.replace('cos://', '').split('/', 1)
                        tile = download(cos_client, bucket, key, tmpdir)
                        tiles.append(tile)
                    if bbox[2] > bbox[0]:
                        merged_image = os.path.join(download_dir, layer, 'merged.tif')
                        if force_bbox:
                            merge(tiles, merged_image, bbox)
                        else:
                            merge(tiles, merged_image)
                        merged_images.append(merged_image)
                    # If the given bbox spans over anti-meridian, we have to merge into two parts. Note that this will
                    # also overwrite the value of force_bbox and set it to true.
                    else:
                        merged_image_1 = os.path.join(download_dir, layer, 'merged_1.tif')
                        merged_image_2 = os.path.join(download_dir, layer, 'merged_2.tif')
                        merge(tiles, merged_image_1, (bbox[0], bbox[1], 180, bbox[3]))
                        merge(tiles, merged_image_2, (-180, bbox[1], bbox[2], bbox[3]))
                        merged_images.extend([merged_image_1, merged_image_2])
        return merged_images
    else:
        tiles = []
        for _, row in df.iterrows():
            layer = row['layer']
            date = row['date']
            url = row['url']
            bucket, key = url.replace('cos://', '').split('/', 1)
            tile = download(cos_client, bucket, key, os.path.join(download_dir, layer, str(date)))
            tiles.append(tile)
        return tiles


def query_tiles(sql_client, cos_client, cos_url, target_cos_url, layers, start_date, end_date, bbox, download_dir,
                search_buffer_in_degrees=2, selected_dates=None, merge_tiles=True, merge_by_date=True, force_bbox=True):

    if isinstance(layers, str):
        layers = [layers]

    if isinstance(selected_dates, str) or isinstance(selected_dates, int):
        selected_dates = [selected_dates]

    # By default, the result will be written into target_cos in CSV format. We disable this and make it write as
    # parquet to improvement performance
    if sql_client.target_cos is not None:
        sql_client.target_cos = None

    expanded_bbox = expand_bbox(bbox, search_buffer_in_degrees)

    if selected_dates is None:
        date_query_part = 'date between {} AND {}'.format(start_date, end_date)
    else:
        date_query_part = 'date in {}'.format(tuple(selected_dates))

    if expanded_bbox[2] >= expanded_bbox[0]:
        lon_query_part = 'lon between {} AND {}'.format(expanded_bbox[0], expanded_bbox[2])
    else:
        lon_query_part = '(lon between {} AND 180 OR lon between -180 AND {})'.format(expanded_bbox[0], expanded_bbox[2])

    query_str = """
    SELECT *
    FROM {0} STORED AS PARQUET
    WHERE layer in {1}
    AND {2}
    AND lat between {3[1]} AND {3[3]}
    AND {4}
    INTO {5} STORED AS parquet
    """.format(cos_url, tuple(layers), date_query_part, expanded_bbox, lon_query_part, target_cos_url)
    # ST_Contains(ST_WKTToSQL('BOUNDINGBOX({} {}, {} {})'), ST_Point(lon, lat))

    df = sql_client.run_sql(query_str)

    if not os.path.exists(download_dir):
        os.makedirs(download_dir)

    if merge_tiles:
        from pyrip.image import merge
        merged_images = []
        for layer in layers:
            tmp_df = df[df['layer'] == layer].sort_values(['date', 'lat', 'lon'])
            if merge_by_date:
                dates = tmp_df['date'].unique()
                for date in dates:
                    tiles = []
                    with tempfile.TemporaryDirectory() as tmpdir:
                        for url in tmp_df[tmp_df['date'] == date]['url']:
                            bucket, key = url.replace('cos://', '').split('/', 1)
                            tile = download(cos_client, bucket, key, tmpdir)
                            tiles.append(tile)
                        if bbox[2] > bbox[0]:
                            merged_image = os.path.join(download_dir, layer, str(date), 'merged.tif')
                            if force_bbox:
                                merge(tiles, merged_image, bbox)
                            else:
                                merge(tiles, merged_image)
                            merged_images.append(merged_image)
                        # If the given bbox spans over anti-meridian, we have to merge into two parts. Note that this
                        # will also overwrite the value of force_bbox and set it to true.
                        else:
                            merged_image_1 = os.path.join(download_dir, layer, str(date), 'merged_1.tif')
                            merged_image_2 = os.path.join(download_dir, layer, str(date), 'merged_2.tif')
                            merge(tiles, merged_image_1, (bbox[0], bbox[1], 180, bbox[3]))
                            merge(tiles, merged_image_2, (-180, bbox[1], bbox[2], bbox[3]))
                            merged_images.extend([merged_image_1, merged_image_2])
            else:
                tiles = []
                with tempfile.TemporaryDirectory() as tmpdir:
                    for url in tmp_df['url']:
                        bucket, key = url.replace('cos://', '').split('/', 1)
                        tile = download(cos_client, bucket, key, tmpdir)
                        tiles.append(tile)
                    if bbox[2] > bbox[0]:
                        merged_image = os.path.join(download_dir, layer, 'merged.tif')
                        if force_bbox:
                            merge(tiles, merged_image, bbox)
                        else:
                            merge(tiles, merged_image)
                        merged_images.append(merged_image)
                    # If the given bbox spans over anti-meridian, we have to merge into two parts. Note that this will
                    # also overwrite the value of force_bbox and set it to true.
                    else:
                        merged_image_1 = os.path.join(download_dir, layer, 'merged_1.tif')
                        merged_image_2 = os.path.join(download_dir, layer, 'merged_2.tif')
                        merge(tiles, merged_image_1, (bbox[0], bbox[1], 180, bbox[3]))
                        merge(tiles, merged_image_2, (-180, bbox[1], bbox[2], bbox[3]))
                        merged_images.extend([merged_image_1, merged_image_2])
        return merged_images
    else:
        tiles = []
        for _, row in df.iterrows():
            layer = row['layer']
            date = row['date']
            url = row['url']
            bucket, key = url.replace('cos://', '').split('/', 1)
            tile = download(cos_client, bucket, key, os.path.join(download_dir, layer, str(date)))
            tiles.append(tile)
        return tiles



import os
import rasterio
from rasterio.warp import calculate_default_transform, Resampling, reproject as rasterio_reproject
from .util import run_command


# def reproject(infile, outfile=None, t_srs='EPSG:4326'):
#     outfile = outfile or os.path.splitext(infile)[0] + '_' + t_srs.replace(':', '') + os.path.splitext(infile)[1]
#     subprocess.run(['gdalwarp', '-t_srs', t_srs,'-overwrite', infile, outfile], check=True, stdout=subprocess.DEVNULL)
#     return outfile


def reproject(infile, outfile=None, dst_crs='EPSG:4326'):
    outfile = outfile or os.path.splitext(infile)[0] + '_' + dst_crs.replace(':', '') + os.path.splitext(infile)[1]
    with rasterio.open(infile) as src:
        with rasterio.Env(CHECK_WITH_INVERT_PROJ=True):
            transform, width, height = calculate_default_transform(
                src.crs, dst_crs, src.width, src.height, *src.bounds)
        kwargs = src.meta.copy()
        kwargs.update({
            'crs': dst_crs,
            'transform': transform,
            'width': width,
            'height': height
        })
        with rasterio.open(outfile, 'w', **kwargs) as dst:
            for i in range(1, src.count + 1):
                rasterio_reproject(
                    source=rasterio.band(src, i),
                    destination=rasterio.band(dst, i),
                    src_transform=src.transform,
                    src_crs=src.crs,
                    dst_transform=transform,
                    dst_crs=dst_crs,
                    resampling=Resampling.nearest)
    return outfile


def warp(infile, outfile, bounds, resolution=None, shape=None, align_pixels=False):
    outfile = outfile or os.path.splitext(infile)[0] + '_warped' + os.path.splitext(infile)[1]
    args = ['gdalwarp']
    args.extend(['-te', *bounds])
    if resolution is None and shape is None:
        raise ValueError('either resolution or shape must to be set.')
    elif resolution is not None and shape is not None:
        raise ValueError('resolution and shape cannot be set at the same time.')
    elif resolution is not None:
        args.extend(['-tr', resolution, resolution])
        if align_pixels:
            args.append('-tap')
    else:
        args.extend(['-ts', shape[1], shape[0]])
    args.extend(['-overwrite', infile, outfile])
    run_command(args)
    return outfile


def align(image, base_image, outfile=None):
    outfile = outfile or os.path.splitext(image)[0] + '_aligned' + os.path.splitext(image)[1]
    with rasterio.open(base_image) as dataset:
        bounds = dataset.bounds
        shape = dataset.shape
    return warp(image, outfile, bounds, shape=shape)




from setuptools import setup, find_packages

version = {}
exec(open('pyrip/version.py').read(), version)
VERSION = version['__version__']

setup(
    name='pyrip',
    version=VERSION,
    description='Raster Image Processor - a thin wrap that provides some convenient functions over gdal',
    packages=find_packages(),
    url='https://github.com/lchu-ibm/pyrip',
    author='Linsong Chu',
    author_email='lchu@us.ibm.com',
    install_requires=['rasterio', 'pandas', 'numpy', 'shapely', 'pyproj']
)



import gzip
import os
import shutil
import tempfile
import zipfile
from pathlib import Path

import numpy as np
import pandas as pd
import rasterio
from osgeo import gdal

from .band import stack_bands
from .util import create_xyz_vrt, run_command

# Fix a bug in conda gdal: https://github.com/OSGeo/gdal/issues/1231
if ';' in os.environ["PATH"]:
    os.environ["PATH"] = os.environ["PATH"].split(';')[1]


def tif_to_numpy(file):
    with rasterio.open(file) as dataset:
        if dataset.count == 1:
            return dataset.read(1)
        else:
            return dataset.read()


def tif_to_df(infile, drop_nodata=True):
    dataset = rasterio.open(infile)
    nrows, ncols = dataset.shape
    # if nrows * ncols > split_size * split_size:
    #     dfs = []
    #     tif_files = split(infile, split_size)
    #     for tif_file in tif_files:
    #         dfs.append(tif_to_df(tif_file, drop_nodata, split_size))
    #     return pd.concat(dfs)
    value = dataset.read(1).ravel()
    a, b, c, d, e, f, g, _, _ = dataset.transform
    affine_matrix = np.array([[a, b, c],
                              [d, e, f],
                              [0, 0, 1]])
    xy_pixel = np.stack((np.tile(np.arange(ncols), nrows) + 0.5,
                         np.repeat(np.arange(nrows), ncols) + 0.5,
                         np.repeat(1, ncols * nrows)))
    if drop_nodata:
        mask = value != dataset.nodata
        value = value[mask]
        xy_pixel = xy_pixel[:, mask]
    xy = np.matmul(affine_matrix, xy_pixel)
    df = pd.DataFrame({'lat': xy[1], 'lon': xy[0], 'value': value})
    # Cast unsigned int (uint8, uint16, uint32) to int32 as spark does not support unsigned int in parquet files
    if 'uint' in dataset.dtypes[0].lower():
        df['value'] = df['value'].astype('int32')
    return df


def xyz_to_tif(infile, outfile, xres, yres=None, bbox=None, lat_col='lat', lon_col='lon', val_col='value', nodata=-9999, dtype='Float64'):
    yres = yres or xres
    vrt = create_xyz_vrt(infile, lat_col, lon_col)
    with tempfile.NamedTemporaryFile(suffix='.vrt') as tmpfile:
        vrt.write(tmpfile.name)
        args = ['gdal_rasterize', tmpfile.name, outfile]
        # layer name, currently hardcoded in util.create_xyz_vrt()
        # args.extend(['-l', os.path.splitext(os.path.basename(infile))[0]])
        # output format
        args.extend(['-of', 'GTiff'])
        # set nodata value
        args.extend(['-a_nodata', nodata])
        # set output data type
        args.extend(['-ot', dtype])
        # attribute column used for burn-in value
        args.extend(['-a', val_col])
        # set target resolution in georeferenced units
        args.extend(['-tr', xres, yres])
        # set output bounding box, leave none to compute boundary's bbox automatically
        if bbox is not None:
            xmin, ymin, xmax, ymax = bbox
            args.extend(['-te', xmin, ymin, xmax, ymax])
        run_command(args)
    return outfile


def df_to_tif(df, outfile, xres, yres=None, bbox=None, lat_col='lat', lon_col='lon', val_col='value', nodata=-9999, dtype='Float64', band_col=None, band_desc_col=None):
    if band_col is None:
        with tempfile.NamedTemporaryFile(suffix='.csv') as tmpfile:
            df.to_csv(tmpfile.name, index=False)
            return xyz_to_tif(tmpfile.name, outfile, xres, yres, bbox, lat_col, lon_col, val_col, nodata, dtype)
    else:
        with tempfile.TemporaryDirectory() as tmpdir:
            bands = sorted(df[band_col].unique())
            tif_files = []
            for band in bands:
                csv_file = os.path.join(tmpdir, 'band_{}.csv'.format(band))
                tif_file = os.path.join(tmpdir, 'band_{}.tif'.format(band))
                tmp_df = df[df[band_col] == band]
                tmp_df.to_csv(csv_file, index=False)
                xyz_to_tif(csv_file, tif_file, xres, yres, bbox, lat_col, lon_col, val_col, nodata, dtype)
                # Set band description
                if band_desc_col is None:
                    band_desc = 'band_{}'.format(band)
                else:
                    band_desc = tmp_df[band_desc_col].iloc[0]
                ds = gdal.Open(tif_file, gdal.GA_Update)
                rb = ds.GetRasterBand(1)
                rb.SetDescription(band_desc)
                del ds
                tif_files.append(tif_file)
            return stack_bands(tif_files, outfile)


def hdf_to_tif(infile, outdir=None, match_substrs=None):
    if infile.endswith('.gz'):
        decompressed_file = os.path.splitext(infile)[0]
        with gzip.open(infile) as infile_:
            with open(decompressed_file, 'wb') as outfile_:
                shutil.copyfileobj(infile_, outfile_)
        return hdf_to_tif(decompressed_file, outdir, match_substrs)
    ds = gdal.Open(infile, gdal.GA_ReadOnly)
    outfiles = []
    for sub_ds in ds.GetSubDatasets():
        sub_ds_name = sub_ds[0]
        sub_ds_layer_name = sub_ds_name.split(':')[-1]
        if match_substrs is not None and not any(match_substr in sub_ds_layer_name for match_substr in match_substrs):
            continue
        basename = os.path.splitext(os.path.basename(infile))[0]
        connector = '_' if basename.count('_') >= basename.count('.') else '.'
        if outdir is None:
            outfile = os.path.splitext(infile)[0] + connector + sub_ds_layer_name + '.tif'
        else:
            if not os.path.exists(outdir):
                os.makedirs(outdir)
            outfilename = os.path.splitext(os.path.basename(infile))[0] + connector + sub_ds_layer_name + '.tif'
            outfile = os.path.join(outdir, outfilename)
        args = ['gdal_translate', '-of', 'GTiff', sub_ds_name, outfile]
        run_command(args)
        outfiles.append(outfile)
    return outfiles


nc_to_tif = hdf_to_tif


def safe_to_tif(infile, outdir=None, patterns=None):
    outdir = outdir or str(Path(infile).resolve().parent)
    patterns = patterns or ['*.jp2']
    # if .SAFE dir:
    if os.path.isdir(infile):
        outfiles = []
        jp2_files = []
        for pattern in patterns:
            jp2_files.extend(Path(infile).rglob(pattern))
        for jp2_file in jp2_files:
            outfile = os.path.join(outdir, os.path.splitext(os.path.basename(str(jp2_file)))[0] + '.tif')
            args = ['gdal_translate', '-of', 'GTiff', str(jp2_file), outfile]
            run_command(args)
            outfiles.append(outfile)
        return outfiles
    elif zipfile.is_zipfile(infile):
        with tempfile.TemporaryDirectory() as tmpdir:
            with zipfile.ZipFile(infile) as zip_ref:
                zip_ref.extractall(tmpdir)
            return safe_to_tif(tmpdir, outdir, patterns)
    else:
        raise TypeError("infile must be either a directory or zip file")


def grib_to_tif(infile, outfile=None):
    outfile = outfile or os.path.splitext(infile)[0] + '.tif'
    args = ['gdal_translate', '-of', 'GTiff', '-ot', 'Byte', '-scale', infile, outfile]
    run_command(args)
    return outfile
